#include "stdafx.h"
#include "clf_util.h"
#include "clf_random.h"
#include "clf_primitives.h"
#include "clf_math.h"
#include "clf_camera.h"
#include "clf_resources.h"
#include "clf_light.h"
#include "clf_font.h"
#include "clf_shader.h"
#include "clf_skybox.h"
#include "clf_effects.h"
#include "clf_quaternion.h"

#include "clf_world.h"

extern HWND  hWnd;						// Window Handle
extern bool keyDown[256];			// holds true / false for each key

static int m_nFPS = 0;
static int m_nFrames = 0;											// FPS and FPS Counter
static bool g_pause=false;

clf_stage g_stage;

// a light source
clf_light g_light(GL_LIGHT0);

// the shader
clf_shader shader;

clf_skybox g_skybox(80);  // size of skybox is small enough for fog effect!!

// class to handle the fog effect
clf_fog g_fog;

// the game world
clf_world g_w;

class clf_ground : public clf_world_object
{
public:
	clf_ground():clf_world_object(1.e30f, 0){
		m_fixed_pos = true;
	}
	virtual ~clf_ground(){
	}
	virtual void render(){
		// glColor4f(.3f, .8f, .0f, 1.f);
		glColor4f(.1f, .1f, .1f, 1.f);
		glBegin(GL_QUADS);
		glNormal3f(0,1,0);
		glVertex3f(-100, 0, -100);
		glVertex3f(-100, 0, 100);
		glVertex3f(100, 0, 100);
		glVertex3f(100, 0, -100);
		glEnd();
	}
	virtual bool collides_with(clf_world_object* pob){
		return false;
	}
};

class building : public clf_world_object
{
public:
	int m_tex;
	clf_vector3D m_size;  // size
	building(const clf_vector3D& p, const clf_vector3D& s):clf_world_object(1.e30f, 0) {
		m_tex = -1;
		m_pos = p; // position
		m_size = s; 
		m_fixed_pos = true;
	}
	virtual ~building(){
	}
	void set_texture(int tex){
		m_tex = tex;
	}
	virtual void render(){
		if (m_tex >= 0)
		{
			clf_enable_textures(m_tex);
			glColor4f(1.f, 1.1f, 1.f, 1.f);
		}
		else
			glColor4f(.7f, .7f, .7f, 1.f);
		glPushMatrix();
		glTranslatef(m_pos.x, m_pos.y, m_pos.z);
		drawBox(m_size.x, m_size.y, m_size.z);
		glPopMatrix();
		if (m_tex >= 0)
			clf_disable_textures();
	}
	virtual bool collides_with(clf_world_object* pob){
		return false;
	}
	virtual bool is_at_xz(clf_vector3D& pos, float r){
		float xmin = m_pos.x-r-m_size.x/2.f;
		float xmax = m_pos.x+r+m_size.x/2.f;
		float zmin = m_pos.z-r-m_size.z/2.f;
		float zmax = m_pos.z+r+m_size.z/2.f;
		if (xmin > pos.x)
			return false;
		if (xmax < pos.x)
			return false;
		if (zmin > pos.z)
			return false;
		if (zmax < pos.z)
			return false;
		return true;
	}
};

class person{
public:
	clf_vector3D m_pos;
	clf_vector3D m_vel;
	clf_vector3D m_look;
	clf_vector3D m_up;
	cCamera m_cam;
	clf_quaternion m_view;
	float direction_angle;
	float horizontal_angle;
	float vertical_angle;
	float m_eye_height;
public:
	person(const clf_vector3D& p) : m_pos(p), m_vel(0,0,5.5), m_look(0,0,1), m_eye_height(1.75f), m_up(0,1,0){
		m_cam.setup(clf_vector3D(m_pos.x,m_pos.y+m_eye_height,m_pos.z),     // position eye level
								clf_vector3D(m_pos.x+m_look.x,m_pos.y+m_eye_height+m_look.y,m_pos.z+m_look.z),        // setup the point of interest (where we are looking at)
								clf_vector3D(m_up.x,m_up.y,m_up.z));       // which way is up
		direction_angle=0;
		horizontal_angle=0;
		vertical_angle=0;
		m_view = make_q_from_euler_angles(vertical_angle,horizontal_angle+direction_angle,0);
	}
	virtual ~person(){
	}
	void apply_view(){
		m_cam.apply();
	}
	void recalc_view(){
		m_view = make_q_from_euler_angles(vertical_angle,horizontal_angle+direction_angle,0);
		m_look = qv_rotate(m_view, clf_vector3D(0,0,1));
		m_up = qv_rotate(m_view, clf_vector3D(0,1,0));
	}
	// all the movements are based on time
	void turn(float fElapsed, int dir){
		if (!dir)
			return;

		float angle = (float)dir*fElapsed*60.f;
		direction_angle += angle;
		if (direction_angle > 360.f)
			direction_angle -= 360.f;
		else if (direction_angle < 0)
			direction_angle = 360.f - direction_angle;

		clf_quaternion rott = make_q_from_euler_angles(0,angle,0);
		m_vel = qv_rotate(rott, m_vel);
		recalc_view();
	}
	void turn_head(float fElapsed, int dir){
		if (!dir)
			return;
		float angle = (float)dir*fElapsed*60.f;
		horizontal_angle += angle;
		if (horizontal_angle > 90.f)
			horizontal_angle = 90.f;
		if (horizontal_angle < -90.f)
			horizontal_angle = -90.f;
		recalc_view();
	}
	void head_ud(float fElapsed, int dir){
		if (!dir)
			return;
		float angle = (float)dir*fElapsed*60.f;
		vertical_angle += angle;
		if (vertical_angle > 90.f)
			vertical_angle = 90.f;
		if (vertical_angle < -90.f)
			vertical_angle = -90.f;
		recalc_view();
	}
	void look_front(){
		horizontal_angle=0;
		vertical_angle=0;
		recalc_view();
	}
	void move(float fElapsed){
		clf_vector3D n_pos = m_pos + m_vel*fElapsed;

		clf_world_object* ob = g_w.object_at(n_pos, .35f);

		m_pos.y = n_pos.y = 0;
		if (ob)
		{
			building* bu = dynamic_cast<building*>(ob);
			if (!bu)
				return;
			float h = bu->m_size.y;
			if (h > 0.5f)
				return;
			n_pos.y = h;
		}
		m_pos = n_pos;

		m_cam.move_to(clf_vector3D(m_pos.x,m_pos.y+m_eye_height,m_pos.z));
		m_cam.look_at(clf_vector3D(m_pos.x+m_look.x,m_pos.y+m_eye_height+m_look.y,m_pos.z+m_look.z));
		m_cam.orient(m_up);
	}
	float x(){
		return m_pos.x;
	}
	float y(){
		return m_pos.y;
	}
	float z(){
		return m_pos.z;
	}
	void render(){
		// render a couple of boxes to look like a car
		glPushMatrix();
		glTranslatef(m_pos.x, m_pos.y, m_pos.z);
		glRotatef(direction_angle, 0,1,0);
		glColor4f(.95f, .95f, .95f, 1.f);
		drawBox(1.7f, 2.8f, 2.2f);
		glColor4f(.4f, .4f, .95f, 1.f);
		drawBox(1.8f, 1.6f, 4.2f);
		glPopMatrix();
	}
};
person *g_person;

// camera that follows the action from a distance
class c_moving_camera : public cCamera{
	clf_vector3D tLocation;      // the tagret location
	clf_vector3D tForward;       // target forward direction
	clf_vector3D tUp;            // target up vector
	clf_vector3D tLookAt;        // target point of interest
	float m_speed;               // camera speed
public:
	c_moving_camera() : cCamera(), tLookAt(0,0,0), m_speed(0.01f) {
	}
	c_moving_camera(cCamera& cp) : cCamera(cp){
	}
	c_moving_camera(c_moving_camera& cp) : cCamera(cp){
	}
	virtual ~c_moving_camera(){}
	// override setup, and use it to set target locations
	virtual void move_to(clf_vector3D& pos){
		tLocation = pos;
	}
	virtual void look_at(clf_vector3D& lookat){
		tLookAt = lookat;
		tForward = tLookAt-tLocation;
		tForward.normalize();
	}
	virtual void orient(clf_vector3D& up){
		tUp = up;
	}
	// actual camera location is set gradualy using camera speed
	void frame_move(float fElapsed){
		clf_vector3D dist(tLocation - vLocation);
		if (dist.length() < 0.001f)
			return;
		// here we set only the camera position
		clf_vector3D pos(vLocation + m_speed*dist);
		cCamera::move_to(pos);
		// and leave the rest
		cCamera::look_at(tLookAt);
		cCamera::vUp = tUp;
	}
};

c_moving_camera g_viewer;

int g_view_type = 0;

void create_world()
{
	clf_ground *g = new clf_ground();
	g_w.add_object(g);

	building *bld;

	int txs[2];
	txs[0] = clf_loadTexture("build1.tga");
	txs[1] = clf_loadTexture("build2.tga");

	float x,z;
	int i=0;
	for (x=-36.f; x<37.f; x+=24.f)
	{
		i = 1-i;
		for (z=-36.f; z<37.f; z+=24.f)
		{
			bld = new building(clf_vector3D(x,5.2f,z), clf_vector3D(10,10,10));
			bld->set_texture(txs[i]);
			g_w.add_object(bld);
			bld = new building(clf_vector3D(x,.1f,z), clf_vector3D(15,0.2f,15));
			g_w.add_object(bld);
			i = 1-i;
		}
	}

	g_person = new person(clf_vector3D(0,0,-75));
}

void destroy_world()
{
	std::list<clf_world_object*>::iterator it;
	for (it=g_w.m_objects.begin(); it!=g_w.m_objects.end(); ++it)
		delete (*it);
}


///////////////////////////////////////////////////////////////////////
void restart_game()
{
}

// this function is called when the program starts to initialize the game
void game_init()
{
	init_math_lib();
	get_font_manager().create_font_2d("Bookman2D", "Bookman Old Style", 16);

	glEnable (GL_DEPTH_TEST);        // enable depth test
	glEnable(GL_COLOR_MATERIAL);     // use color for drawing the object

	// load shader
	shader.load_shaders("phong_vertex_shader2.glsl", "fragment_shader1.glsl");
	
	// Light model parameters:
	// -------------------------------------------
	// create and position a light source
	g_light.set_ambient(0.25f, 0.25f, 0.25f, 1.f);
	g_light.set_diffuse(0.7f, 0.7f, 0.7f, 1.f);
	g_light.set_specular(1.f, 1.f, 1.f, 1.f);
	g_light.set_position(-100.f, 100.f, 100.f, 0.f);

	// -------------------------------------------
	// Material parameters:
	
	GLfloat material_ambient[] = {0.7f, 0.7f, 0.7f, 1.0f};
	GLfloat material_diffuse[] = {0.4f, 0.4f, 0.4f, 1.0f};
	GLfloat material_specular[] = {0.8f, 0.8f, 0.8f, 1.0f};
	GLfloat material_emission[] = {0.1f, 0.1f, 0.1f, 0.0f};
	GLfloat material_shine = 50.0f;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, material_ambient);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, material_diffuse);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, material_specular);
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, material_emission);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, material_shine);

	g_skybox.load_textures("skybox/up.tga", "skybox/dn.tga", "skybox/ft.tga", "skybox/bk.tga", "skybox/lt.tga", "skybox/rt.tga");

	create_world();

	restart_game();
}

// this function is called when the window is resized
void resize_window (int width, int height)			// reshape the window when it's moved or resized
{
	// setup viewport
	g_stage.resize(0,0,width, height);
}

// this function is called upon exit to clear allocated resources
void game_terminate()
{
	destroy_world();
	shader.terminate();
}

// the core of the game. This is called repeatedly to update the state of the game
// fElapsed: is the elapsed time since last call. The time is in seconds (usualy a very small number)
void frame_move(float fElapsed)
{
	if (keyDown[VK_ESCAPE]) 
		DestroyWindow(hWnd);
	if (keyDown['R'])
	{
		keyDown['R'] = 0;
		restart_game();
	}
	if (keyDown['P'])
	{
		keyDown['P'] = 0;
		g_pause = !g_pause;
	}

	if (keyDown['V'])
	{
		keyDown['V'] = 0;
		g_view_type = 1 - g_view_type;
	}
	

	int dir = 0;
	if (keyDown['Z'])
		dir = -1;
	else if (keyDown['X'])
		dir = 1;

	int dv = 0;
	if (keyDown['N'])
		dv = -1;
	else if (keyDown['M'])
		dv = 1;

	int du = 0;
	if (keyDown['H'])
		du = -1;
	else if (keyDown['B'])
		du = 1;

	if (!g_pause)
	{
		g_person->turn(fElapsed, dir);
		g_person->turn_head(fElapsed, dv);
		g_person->head_ud(fElapsed, du);
		if (keyDown['J'])
			g_person->look_front();
		g_person->move(fElapsed);
		g_w.step_simulation(fElapsed);

		// position the moving camera
		// 1. get moving object velocity vector
		clf_vector3D s(g_person->m_vel);
		// and make it unit length
		s.normalize();
		// 2. create the view distance vector
		s.scale(15.f);
		// 3. and calculate the camera position
		clf_vector3D loc(g_person->m_pos - s);
		// raise the view camera at a certain height
		loc.y = 2.f;
		// apply all these to the moving camera
		g_viewer.move_to(loc);
		g_viewer.look_at(g_person->m_pos);
		g_viewer.orient(clf_vector3D(0,1,0));
		// and let it move
		g_viewer.frame_move(fElapsed);
	}

	// frame counting mechanism
	// the higher the frame rate the faster the system
	// if frame rate drops below 30 we are in deep trouble
	// we should either optimize the program or buy a new computer
	static dFloat tot = 0;
	tot += fElapsed;
	m_nFrames++;												// Increment Our frame Counter
	if (tot >= 1.f)
	{
		tot = 0;
		m_nFPS = m_nFrames;
		m_nFrames = 0;

		char txt[200];
		sprintf(txt, "GusOnGames, fps:%d",m_nFPS);
		SetWindowText(hWnd, txt);
	}
}

void frame_render()
{
	// clear the screen
	// drawing the skybox means we do not have to waste time clearing the screen
	glClear(/*GL_COLOR_BUFFER_BIT |*/ GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);	// clear screen and stencil buffer

	glMatrixMode (GL_PROJECTION);									// select the projection matrix
	glLoadIdentity ();														// reset the projection matrix

	// viewport and perspective
	g_stage.prepare();

	glMatrixMode (GL_MODELVIEW);								// select the modelview matrix
	glLoadIdentity ();													// reset

	glPushMatrix();

	// position the viewer
	if (g_view_type == 0)
		g_viewer.apply();
	else
		g_person->apply_view();

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glEnable(GL_NORMALIZE);
	// position the light
	g_light.enable();

	// fog
	g_fog.fog_apply();
	// render the skybox
	g_skybox.render(g_person->x(), g_person->y(), g_person->z());

	// shader
	shader.begin();

	// reapply the light
	g_light.apply();

	if (g_view_type == 0)
		g_person->render();

	// draw world objects
	g_w.render();

	shader.end();
	glPopMatrix();
	glPopAttrib();
}
